#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <unistd.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/wait.h>
#include <stdio.h>

static char** create( const char *s ) {
  char *p = (char*)s, *f;
  int n, i;
  for( n = 0; ; n++ ) {
    if( ( p = strpbrk( p, " " ) ) == NULL ) break;
    while( *p == ' ' ) p++;
  };
  char **pv = malloc( ( n + 2 ) * sizeof( char* ) );
  for( i = 0; i < n + 2; i++ ) pv[ i ] = NULL;
  p = (char*)s;
  f = strpbrk( p, " " );
  for( n = 0; ; n++ ) {
    int k = ( f - p );
    pv[ n ] = malloc( k + 1 );
    strncpy( pv[ n ], p, k );
    pv[ n ][ k ] = '\0';
    p = f;
    while( *p == ' ' ) p++;
    if( ( f = strpbrk( p, " " ) ) == NULL ) {
      pv[ n + 1 ] = strdup( p ); 
      break;
    }
  }
  return pv;
};

static int debug_level = 0;
static u_long buflen = 1024;
static u_char *buf;
static int direct = 0;

int main( int argc, char *argv[] ) {
   int c;
   double stret = 1.0;
   while( -1 != ( c = getopt( argc, argv, "b:edvh" ) ) )
      switch( c ) {
         case 'b': if( 0 != atol( optarg ) ) buflen = atol( optarg ); break;
         case 'e': direct = +1; break; // кодировать
         case 'd': direct = -1; break; // декодировать
         case 'v': debug_level++; break;
         case 'h':
         default : printf( "%s [<опции>] <имя вх.файла> <имя вых.файла>\n"
                           "опции:\n"
                           " -b - размер аудио буфера\n"
                           " -e - raw - > speex\n"
                           " -d - speex - > raw\n"                     
                           " -v - увеличить уровень детализации отладочного вывода\n"
                           " -h - вот этот текст подсказки\n",
                           argv[ 0 ] );
            exit( 'h' == c ? EXIT_SUCCESS : EXIT_FAILURE );
      }
   if( 0 == direct ) printf( "должно быть указано направление преобразования: '-e' или '-d'\n" ), exit( EXIT_FAILURE );
   if( optind != argc - 2 ) printf( "должно быть указаны имена входного и выходного звуковых файлов\n" ), exit( EXIT_FAILURE ); 
   // файл с которого читается входной аудиопоток
   int fai = open( argv[ optind ], O_RDONLY );
   if( -1 == fai ) perror( "open input" ), exit( EXIT_FAILURE );
   // файл в который пишется результирующий аудиопоток
   int fao = open( argv[ optind + 1 ], O_WRONLY | O_CREAT | O_TRUNC,
                                       S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP | S_IROTH | S_IWOTH );
   if( -1 == fao ) perror( "open output" ), exit( EXIT_FAILURE );
   const char *fmtstr[] = { "speexenc --rate 8000 --16bit %s - -", "speexdec --rate 8000 --mono %s - -" };
   char comstr[ 80 ];
#define DEBUG_BITRATE 3                     
   sprintf( comstr, fmtstr[ ( direct > 0 ? 0 : 1 ) ], ( debug_level >= DEBUG_BITRATE ? "-V" : "" ) );
// --comp n 
// Set encoding complexity (0-10), default 3
// --quality n
// Encoding quality (0-10), default 3

   int fi[ 2 ], // pipe – для ввода в дочернем процессе - для вывода из main;
       fo[ 2 ]; // pipe – для вывода в дочернем процессе - для ввода в main;
   if( pipe( fi ) || pipe( fo ) ) perror( "pipe" ), exit( EXIT_FAILURE );
   pid_t pid = fork();
   switch( pid ) {
      case -1: perror( "fork" ), exit( EXIT_FAILURE );
      case  0:   // дочерний клон
         close( fi[ 1 ] ), close( fo[ 0 ] );
         if( dup2( fi[ 0 ], STDIN_FILENO ) == -1 ||
             dup2( fo[ 1 ], STDOUT_FILENO ) == -1 ) perror( "dup2" ), exit( EXIT_FAILURE );
         close( fi[ 0 ] ), close( fo[ 1 ] );
         if( 0 == debug_level )  // запретить вывод STDERR
             if( dup2( open( "/dev/null", O_WRONLY ), STDERR_FILENO ) == -1 ) perror( "stderr dup2" ), exit( EXIT_FAILURE );
         // здесь создаётся список параметров запуска
         char **pv = create( comstr ); 
         // запуск консольного клиента         
         if( -1 == execvp( pv[ 0 ], pv ) ) perror( "execvp" ), exit( EXIT_FAILURE );
         // ... возврата отсюда уже не будет ...
      default:   // родительский процесс
         close( fi[ 0 ] ), close( fo[ 1 ] );
         int cur_flg;
         // чтение из родительского процесса должно быть в режиме O_NONBLOCK
         cur_flg = fcntl( fo[ 0 ], F_GETFL ); 
         if( -1 == fcntl( fo[ 0 ], F_SETFL, cur_flg | O_NONBLOCK ) )
            perror( "fcntl" ), exit( EXIT_FAILURE );
         // для записи O_NONBLOCK не обязательно
         break;
   };
   int fdi = fo[ 0 ], fdo = fi[ 1 ];
   // дескриптор куда записывается то, что читает из stdin дочерний процесс
   if( -1 == fdo ) perror( "pipe output" ), exit( EXIT_FAILURE );
   // дескриптор с котрого читается вывод в stdout дочернего процесса
   if( -1 == fdi ) perror( "pipe input" ), exit( EXIT_FAILURE );
   buf = malloc( buflen );
#define DEBUG_IN_PROC 2  
   int sum[] = { 0, 0, 0, 0 };
   while( 1 ) {
      int n;
      if( fai > 0 ) {
         n = read( fai, buf, buflen );
         sum[ 0 ] += n > 0 ? n : 0;
         if( debug_level >= DEBUG_IN_PROC ) printf( "READ from in file\t%d -> %d\n", n, sum[ 0 ] );
         if( -1 == n ) perror( "read file" ), exit( EXIT_FAILURE );
         if( 0 == n ) close( fai ), fai = -1;
      };
      if( fai > 0 ) {
         n = write( fdo, buf, n );
         sum[ 1 ] += n > 0 ? n : 0;
         if( debug_level >= DEBUG_IN_PROC ) printf( "WRITE to child\t%d -> %d\n", n, ( sum[ 1 ] += n > 0 ? n : 0 ) );
         if( -1 == n ) perror( "write pipe" ), exit( EXIT_FAILURE );
         // передеспетчеризация - дать время на обработку
         usleep( 100 );
      }
      else close( fdo ), fdo = -1;
      n = read( fdi, buf, buflen );
      if( debug_level >= DEBUG_IN_PROC ) {
         printf( "READ from child\t%d -> %d", n, ( sum[ 2 ] += n > 0 ? n : 0 ) );
         fflush( stdout );
      }
      if( n >= 0 && debug_level >= DEBUG_IN_PROC ) printf( "\n" );
      // это может быть только после закрытия fdo!!!
      if( 0 == n ) break;
      else if( -1 == n ) {
         if( EAGAIN == errno ) {
            if( debug_level >= DEBUG_IN_PROC ) printf( " : == not ready == ... wait ...\n" );
            usleep( 300 );
            continue;
         }
         else perror( "\nread pipe" ), exit( EXIT_FAILURE );
      }
      n = write( fao, buf, n );
      sum[ 3 ] += n > 0 ? n : 0;      
      if( debug_level >= DEBUG_IN_PROC ) printf( "WRITE to out file\t%d -> %d\n", n, sum[ 3 ] );
      if( -1 == n ) perror( "write file" ), exit( EXIT_FAILURE );
   };
   if( debug_level > 0 ) printf( "summary: read from in %d, write to out %d [bytes]\n", sum[ 0 ], sum[ 3 ] );
   free( buf );
   close( fai ), close( fao );
   close( fdi ), close( fdo );
   wait( NULL );
   return EXIT_SUCCESS;
};
  
