#include "common.h"

static int debug_level = 0;

//void comment_init(char **comments, int* length, char *vendor_string);
//void comment_add(char **comments, int* length, char *tag, char *val);


//Write an Ogg page to a file pointer
int oe_write_page(ogg_page *page, FILE *fp)
{
   int written;
   written = fwrite(page->header,1,page->header_len, fp);
   written += fwrite(page->body,1,page->body_len, fp);
   
   return written;
}

#define MAX_FRAME_SIZE 2000
#define MAX_FRAME_BYTES 2000

// Convert input audio bits, endians and channels 
static int read_samples(FILE *fin, int frame_size, int bits, int channels, int lsb, short * input, char *buff, int *size)
{   
   unsigned char in[MAX_FRAME_BYTES*2];
   int i;
   short *s;
   int nb_read;

   if (size && *size<=0)
   {
      return 0;
   }
   //Read input audio
   if (size)
      *size -= bits/8*channels*frame_size;
   if (buff)
   {
      for (i=0;i<12;i++)
         in[i]=buff[i];
      nb_read = fread(in+12,1,bits/8*channels*frame_size-12, fin) + 12;
      if (size)
         *size += 12;
   } else {
      nb_read = fread(in,1,bits/8*channels* frame_size, fin);
   }
   nb_read /= bits/8*channels;

   //fprintf (stderr, "%d\n", nb_read);
   if (nb_read==0)
      return 0;

   s=(short*)in;
   if(bits==8)
   {
      // Convert 8->16 bits 
      for(i=frame_size*channels-1;i>=0;i--)
      {
         s[i]=(in[i]<<8)^0x8000;
      }
   } else
   {
      // convert to our endian format 
      for(i=0;i<frame_size*channels;i++)
      {
         if(lsb) 
            s[i]=le_short(s[i]); 
         else
            s[i]=be_short(s[i]);
      }
   }

   // FIXME: This is probably redundent now 
   // copy to float input buffer 
   for (i=0;i<frame_size*channels;i++)
   {
      input[i]=(short)s[i];
   }

   for (i=nb_read*channels;i<frame_size*channels;i++)
   {
      input[i]=0;
   }


   return nb_read;
}

void decoder( FILE* fin, FILE* fout, int quiet ) { 
#define MAX_FRAME_SIZE 2000
     int option_index = 0;
     short out[MAX_FRAME_SIZE];
     short output[MAX_FRAME_SIZE];
     int frame_size=0;
     void *st=NULL;
     int packet_count=0;
     int stream_init = 0;
     ogg_int64_t page_granule=0, last_granule=0;
     int skip_samples=0, page_nb_packets;
     ogg_page       og;
     ogg_packet     op;
     ogg_stream_state os;
     int enh_enabled = 1;
     int nframes=2;
     int print_bitrate=0;
     int eos=0;
     int forceMode=-1;
     int audio_size=0;
     float loss_percent=-1;
     SpeexStereoState stereo = SPEEX_STEREO_STATE_INIT;
     int channels=-1;
     int rate=0;
     int extra_headers;
     int wav_format=0;
     int count = 0;

     //Init Ogg data struct
     ogg_sync_state oy;   
     ogg_sync_init(&oy);
     SpeexBits bits;   
     speex_bits_init(&bits);
     int nloop = 0, nsum = 0;
     //Main decoding loop
     while (1) {
         char *data;
         int i = 0, j, nb_read;
         nloop++;
         //Get the ogg buffer for writing
         data = ogg_sync_buffer(&oy, 200);
         //Read bitstream from input file
         nb_read = fread(data, sizeof(char), 200, fin);               
         if(debug_level > 0 ) fprintf (stderr, "loop #%d read from input %d\t=\t%d bytes\n", nloop, nb_read, nsum += nb_read);
         ogg_sync_wrote(&oy, nb_read);
         //Loop for all complete pages we got (most likely only one)
         while (ogg_sync_pageout(&oy, &og) == 1) {
             int packet_no;
             if (stream_init == 0) {
                 ogg_stream_init(&os, ogg_page_serialno(&og));
                 stream_init = 1;
             }
             //Add page to the bitstream
             ogg_stream_pagein(&os, &og);
             page_granule = ogg_page_granulepos(&og);
             page_nb_packets = ogg_page_packets(&og);
             if (page_granule>0 && frame_size) {
                 skip_samples = page_nb_packets*frame_size*nframes - (page_granule-last_granule);
                 if (ogg_page_eos(&og)) skip_samples = -skip_samples;
             //  else if (!ogg_page_bos(&og)) skip_samples = 0;
             } else skip_samples = 0;
             //printf ("page granulepos: %d %d %d\n", skip_samples, page_nb_packets, (int)page_granule);
             last_granule = page_granule;
             //Extract all available packets
             packet_no=0;
             while (!eos && ogg_stream_packetout(&os, &op)==1) {
                 if (packet_count==0) {
                     //If first packet, process as Speex header                 
                     st = process_header(&op, enh_enabled, &frame_size, &rate, &nframes, forceMode, &channels, &stereo, &extra_headers, quiet);
                     if (!nframes) nframes=1;
                     if (!st) exit(1);
   		     write_wav_header(fout, rate, channels, 0, 0);
                 } else if (packet_count==1) { //!!!!!!!!!!!!!!!!!!!!!
                     if (!quiet) print_comments((char*)op.packet, op.bytes);
                 } else if (packet_count<=1+extra_headers) { // Ignore extra headers 
                 } else {
                     int lost=0;
                     packet_no++;
                     if (loss_percent>0 && 100*((float)rand())/RAND_MAX<loss_percent) lost=1;
                     //End of stream condition
                     if (op.e_o_s) eos=1;
                     //Copy Ogg packet to Speex bitstream
                     speex_bits_read_from(&bits, (char*)op.packet, op.bytes);
                     for (j=0;j!=nframes;j++) {
                         int ret;
                         //Decode frame
                         if (!lost) ret = speex_decode_int(st, &bits, output);
                         else ret = speex_decode_int(st, NULL, output);
                         // for (i=0;i<frame_size*channels;i++) printf ("%d\n", (int)output[i]);
                         if (ret==-1) break;
                         if (ret==-2) {
                             fprintf (stderr, "Decoding error: corrupted stream?\n");
                             break;
                         }
                         if (speex_bits_remaining(&bits)<0) {
                             fprintf (stderr, "Decoding overflow: corrupted stream?\n");
                             break;
                         }
                         if (channels==2) speex_decode_stereo_int(output, frame_size, &stereo);
                         if (print_bitrate) {
                             int tmp;
                             char ch=13;
                             speex_decoder_ctl(st, SPEEX_GET_BITRATE, &tmp);
                             fputc (ch, stderr);
                             fprintf (stderr, "Bitrate is use: %d bps     ", tmp);
                         }
                         // Save to output file
                         for (i=0;i<frame_size*channels;i++) out[i]=output[i];
                         int frame_offset = 0;
                         int new_frame_size = frame_size;
                         //printf ("packet %d %d\n", packet_no, skip_samples);
                         if (packet_no == 1 && j==0 && skip_samples > 0) {
                             //printf ("chopping first packet\n");
                             new_frame_size -= skip_samples;
                             frame_offset = skip_samples;
                         }
                         if (packet_no == page_nb_packets && skip_samples < 0) {
                             int packet_length = nframes*frame_size+skip_samples;
                             new_frame_size = packet_length - j*frame_size;
                             if (new_frame_size<0) new_frame_size = 0;
                             if (new_frame_size>frame_size) new_frame_size = frame_size;
                             //printf ("chopping end: %d %d %d\n", new_frame_size, packet_length, packet_no);
                         }
                         if (new_frame_size) {  
                             fwrite(out+frame_offset*channels, sizeof(short), new_frame_size*channels, fout);
                             if(debug_level > 0 )
                                 fprintf (stderr, "write to array %d - frame_size %d - write to output file %d\n",
                                                  i, frame_size, (2*new_frame_size*channels));
                             audio_size += sizeof(short) * new_frame_size * channels;
                         }
                     }
                 }
                 packet_count++;
             }
         }
         if (feof(fin)) break;
     }
     if (wav_format) { // ??????????????????????
         if (fseek(fout, 4, SEEK_SET) == 0) {
             int tmp = le_int(audio_size + 36);
             fwrite(&tmp, 4, 1, fout);
             if (fseek(fout, 32, SEEK_CUR) == 0) {
                 tmp = le_int(audio_size);
                 fwrite(&tmp, 4, 1, fout);
             } else fprintf (stderr, "First seek worked, second didn't\n");
         } else fprintf (stderr, "Cannot seek on wave file, size will be incorrect\n");
     }
     if (st) speex_decoder_destroy(st);
     else fprintf (stderr, "This doesn't look like a Speex file\n");
     speex_bits_destroy(&bits);
     if (stream_init) ogg_stream_clear(&os);
     ogg_sync_clear(&oy);
}


void encoder( FILE* fin, FILE* fout, int quiet ) { 

   int nb_samples, total_samples=0, nb_encoded;
   int c;
   int option_index = 0;
   char *inFile, *outFile;
//   FILE *fin, *fout;
   short input[MAX_FRAME_SIZE];
   int frame_size;
//   int quiet=0;
   int vbr_enabled=0;
   int abr_enabled=0;
   int vad_enabled=0;
   int dtx_enabled=0;
   int nbBytes;
   SpeexMode *mode=NULL;
   int modeID = -1;
   void *st;
   SpeexBits bits;
   char cbits[MAX_FRAME_BYTES];
   int print_bitrate=0;
   int rate=0, size;
   int chan=1;
   int fmt=16;
   int quality=-1;
   float vbr_quality=-1;
   int lsb=1;
   ogg_stream_state os;
   ogg_page 		 og;
   ogg_packet 		 op;
   int bytes_written=0, ret, result;
   int id=-1;
   SpeexHeader header;
   int nframes=1;
   int complexity=3;
//   char *vendor_string = "Encoded with Speex " SPEEX_VERSION;
   char *comments;
   int comments_length;
   int close_in=0, close_out=0;
   int eos=0;
   int bitrate=0;
   double cumul_bits=0, enc_frames=0;
   char first_bytes[12];
   int wave_input=0;
   int tmp;
   int lookahead = 0;
   char *vendor_string = "Encoded with Speex"; // SPEEX_VERSION;

   comment_init(&comments, &comments_length, vendor_string); 
   //Initialize Ogg stream struct
   srand(time(NULL));
   if (ogg_stream_init(&os, rand())==-1)
   {
      fprintf(stderr,"Error: stream init failed\n");
      exit(1);
   }
   {
      fread(first_bytes, 1, 12, fin);
      if (strncmp(first_bytes,"RIFF",4)==0 && strncmp(first_bytes,"RIFF",4)==0)
      {
         if (read_wav_header(fin, &rate, &chan, &fmt, &size)==-1)
            exit(1);
         wave_input=1;
         lsb=1; /* CHECK: exists big-endian .wav ?? */ 
      }
   }

   if (modeID==-1 && !rate)
   {
      /* By default, use narrowband/8 kHz */
      modeID = SPEEX_MODEID_NB;
      rate=8000;
   } else if (modeID!=-1 && rate)
   {
      if (rate>48000)
      {
         fprintf (stderr, "Error: sampling rate too high: %d Hz, try down-sampling\n", rate);
         exit(1);
      } else if (rate>25000)
      {
         if (modeID != SPEEX_MODEID_UWB)
         {
            fprintf (stderr, "Warning: Trying to encode in %s at %d Hz. I'll do it but I suggest you try ultra-wideband instead\n", mode->modeName , rate);
         }
      } else if (rate>12500)
      {
         if (modeID != SPEEX_MODEID_WB)
         {
            fprintf (stderr, "Warning: Trying to encode in %s at %d Hz. I'll do it but I suggest you try wideband instead\n", mode->modeName , rate);
         }
      } else if (rate>=6000)
      {
         if (modeID != SPEEX_MODEID_NB)
         {
            fprintf (stderr, "Warning: Trying to encode in %s at %d Hz. I'll do it but I suggest you try narrowband instead\n", mode->modeName , rate);
         }
      } else {
         fprintf (stderr, "Error: sampling rate too low: %d Hz\n", rate);
         exit(1);
      }
   } else if (modeID==-1)
   {
      if (rate>48000)
      {
         fprintf (stderr, "Error: sampling rate too high: %d Hz, try down-sampling\n", rate);
         exit(1);
      } else if (rate>25000)
      {
         modeID = SPEEX_MODEID_UWB;
      } else if (rate>12500)
      {
         modeID = SPEEX_MODEID_WB;
      } else if (rate>=6000)
      {
         modeID = SPEEX_MODEID_NB;
      } else {
         fprintf (stderr, "Error: Sampling rate too low: %d Hz\n", rate);
         exit(1);
      }
   } else if (!rate)
   {
      if (modeID == SPEEX_MODEID_NB)
         rate=8000;
      else if (modeID == SPEEX_MODEID_WB)
         rate=16000;
      else if (modeID == SPEEX_MODEID_UWB)
         rate=32000;
   }

   if (!quiet)
      if (rate!=8000 && rate!=16000 && rate!=32000)
         fprintf (stderr, "Warning: Speex is only optimized for 8, 16 and 32 kHz. It will still work at %d Hz but your mileage may vary\n", rate); 

   mode = speex_lib_get_mode (modeID);

   speex_init_header(&header, rate, 1, mode);
   header.frames_per_packet=nframes;
   header.vbr=vbr_enabled;
   header.nb_channels = chan;

   {
      char *st_string="mono";
      if (chan==2)
         st_string="stereo";
      if (!quiet)
         fprintf (stderr, "Encoding %d Hz audio using %s mode (%s)\n", 
               header.rate, mode->modeName, st_string);
   }
   /*fprintf (stderr, "Encoding %d Hz audio at %d bps using %s mode\n", 
     header.rate, mode->bitrate, mode->modeName); */

printf("Initialize Speex encoder\n");
   /*Initialize Speex encoder */
   st = speex_encoder_init(mode);

/*   if (strcmp(outFile,"-")==0)
   {
#if defined WIN32 || defined _WIN32
      _setmode(_fileno(stdout), _O_BINARY);
#endif
      fout=stdout;
   }
   else 
   {
      fout = fopen(outFile, "wb");
      if (!fout)
      {
         perror(outFile);
         exit(1);
      }
      close_out=1;
   } */

   speex_encoder_ctl(st, SPEEX_GET_FRAME_SIZE, &frame_size);
   speex_encoder_ctl(st, SPEEX_SET_COMPLEXITY, &complexity);
   speex_encoder_ctl(st, SPEEX_SET_SAMPLING_RATE, &rate);

   if (quality >= 0)
   {
      if (vbr_enabled)
         speex_encoder_ctl(st, SPEEX_SET_VBR_QUALITY, &vbr_quality);
      else
         speex_encoder_ctl(st, SPEEX_SET_QUALITY, &quality);
   }
   if (bitrate)
   {
      if (quality >= 0 && vbr_enabled)
         fprintf (stderr, "Warning: --bitrate option is overriding --quality\n");
      speex_encoder_ctl(st, SPEEX_SET_BITRATE, &bitrate);
   }
   if (vbr_enabled)
   {
      tmp=1;
      speex_encoder_ctl(st, SPEEX_SET_VBR, &tmp);
   } else if (vad_enabled)
   {
      tmp=1;
      speex_encoder_ctl(st, SPEEX_SET_VAD, &tmp);
   }
   if (dtx_enabled)
      speex_encoder_ctl(st, SPEEX_SET_DTX, &tmp);
   if (dtx_enabled && !(vbr_enabled || abr_enabled || vad_enabled))
   {
      fprintf (stderr, "Warning: --dtx is useless without --vad, --vbr or --abr\n");
   } else if ((vbr_enabled || abr_enabled) && (vad_enabled))
   {
      fprintf (stderr, "Warning: --vad is already implied by --vbr or --abr\n");
   }

   if (abr_enabled)
   {
      speex_encoder_ctl(st, SPEEX_SET_ABR, &abr_enabled);
   }

   speex_encoder_ctl(st, SPEEX_GET_LOOKAHEAD, &lookahead);
   
    //Write header
printf("Write header\n");   
   {

      op.packet = (unsigned char *)speex_header_to_packet(&header, (int*)&(op.bytes));
      op.b_o_s = 1;
      op.e_o_s = 0;
      op.granulepos = 0;
      op.packetno = 0;
      ogg_stream_packetin(&os, &op);
      free(op.packet);

      op.packet = (unsigned char *)comments;
      op.bytes = comments_length;
      op.b_o_s = 0;
      op.e_o_s = 0;
      op.granulepos = 0;
      op.packetno = 1;
      ogg_stream_packetin(&os, &op);
      
      while((result = ogg_stream_flush(&os, &og)))
      {
         if(!result) break;
         ret = oe_write_page(&og, fout);
         if(ret != og.header_len + og.body_len)
         {
            fprintf (stderr,"Error: failed writing header to output stream\n");
            exit(1);
         }
         else
            bytes_written += ret;
      }
   }

   free(comments);

   speex_bits_init(&bits);

   if (!wave_input)
   {
      nb_samples = read_samples(fin,frame_size,fmt,chan,lsb,input, first_bytes, NULL);
   } else {
      nb_samples = read_samples(fin,frame_size,fmt,chan,lsb,input, NULL, &size);
   }
   if (nb_samples==0)
      eos=1;
   total_samples += nb_samples;
   nb_encoded = -lookahead;
   /*Main encoding loop (one frame per iteration) */
printf("Main encoding loop\n");
   while (!eos || total_samples>nb_encoded)
   {

static int nloop = 0;
printf("----------------------- loop #%d ------------------\n", ++nloop);
      id++;
      /*Encode current frame */
      if (chan==2)
         speex_encode_stereo_int(input, frame_size, &bits);

      speex_encode_int(st, input, &bits);
      
      nb_encoded += frame_size;
      if (print_bitrate) {
         int tmp;
         char ch=13;
         speex_encoder_ctl(st, SPEEX_GET_BITRATE, &tmp);
         fputc (ch, stderr);
         cumul_bits += tmp;
         enc_frames += 1;
         if (!quiet)
         {
            if (vad_enabled || vbr_enabled || abr_enabled)
               fprintf (stderr, "Bitrate is use: %d bps  (average %d bps)   ", tmp, (int)(cumul_bits/enc_frames));
            else
               fprintf (stderr, "Bitrate is use: %d bps     ", tmp);
         }
         
      }

      if (wave_input)
      {
         nb_samples = read_samples(fin,frame_size,fmt,chan,lsb,input, NULL, &size);
      } else {
         nb_samples = read_samples(fin,frame_size,fmt,chan,lsb,input, NULL, NULL);
      }
      if (nb_samples==0)
      {
         eos=1;
      }
      if (eos && total_samples<=nb_encoded)
         op.e_o_s = 1;
      else
         op.e_o_s = 0;
      total_samples += nb_samples;

      if ((id+1)%nframes!=0)
         continue;
      speex_bits_insert_terminator(&bits);
      nbBytes = speex_bits_write(&bits, cbits, MAX_FRAME_BYTES);
      speex_bits_reset(&bits);
      op.packet = (unsigned char *)cbits;
      op.bytes = nbBytes;
      op.b_o_s = 0;
      /*Is this redundent? */
      if (eos && total_samples<=nb_encoded)
         op.e_o_s = 1;
      else
         op.e_o_s = 0;
      op.granulepos = (id+1)*frame_size-lookahead;
      if (op.granulepos>total_samples)
         op.granulepos = total_samples;
      /*printf ("granulepos: %d %d %d %d %d %d\n", (int)op.granulepos, id, nframes, lookahead, 5, 6);*/
      op.packetno = 2+id/nframes;
      ogg_stream_packetin(&os, &op);

      //Write all new pages (most likely 0 or 1)
      while (ogg_stream_pageout(&os,&og))
      {
         ret = oe_write_page(&og, fout);
         if(ret != og.header_len + og.body_len)
         {
            fprintf (stderr,"Error: failed writing header to output stream\n");
            exit(1);
         }
         else
            bytes_written += ret;
      }
   }
   if ((id+1)%nframes!=0)
   {
      while ((id+1)%nframes!=0)
      {
         id++;
         speex_bits_pack(&bits, 15, 5);
      }
      nbBytes = speex_bits_write(&bits, cbits, MAX_FRAME_BYTES);
      op.packet = (unsigned char *)cbits;
      op.bytes = nbBytes;
      op.b_o_s = 0;
      op.e_o_s = 1;
      op.granulepos = (id+1)*frame_size-lookahead;
      if (op.granulepos>total_samples)
         op.granulepos = total_samples;

      op.packetno = 2+id/nframes;
      ogg_stream_packetin(&os, &op);
   }
   //Flush all pages left to be written
   while (ogg_stream_flush(&os, &og))
   {
      ret = oe_write_page(&og, fout);
      if(ret != og.header_len + og.body_len)
      {
         fprintf (stderr,"Error: failed writing header to output stream\n");
         exit(1);
      }
      else
         bytes_written += ret;
   }

   speex_encoder_destroy(st);
   speex_bits_destroy(&bits);
   ogg_stream_clear(&os);

   if (close_in)
      fclose(fin);
   if (close_out)
      fclose(fout);
   return; // 0;
}

int main(int argc, char **argv) {

   int nb_samples, total_samples=0, nb_encoded;
//   int c;
   int option_index = 0;
   char *inFile, *outFile;
//   FILE *fin, *fout;
   short input[MAX_FRAME_SIZE];
   int frame_size;
   int quiet=0;
   int vbr_enabled=0;
   int abr_enabled=0;
   int vad_enabled=0;
   int dtx_enabled=0;
   int nbBytes;
   SpeexMode *mode=NULL;
   int modeID = -1;
   void *st;
   SpeexBits bits;
   char cbits[MAX_FRAME_BYTES];
   int print_bitrate=0;
   int rate=0, size;
   int chan=1;
   int fmt=16;
   int quality=-1;
   float vbr_quality=-1;
   int lsb=1;
   ogg_stream_state os;
   ogg_page 		 og;
   ogg_packet 		 op;
   int bytes_written=0, ret, result;
   int id=-1;
   SpeexHeader header;
   int nframes=1;
   int complexity=3;

   char *comments;
   int comments_length;
   int close_in=0, close_out=0;
   int eos=0;
   int bitrate=0;
   double cumul_bits=0, enc_frames=0;
   char first_bytes[12];
   int wave_input=0;
   int tmp;
   int lookahead = 0;


     int c;
     while( -1 != ( c = getopt( argc, argv, "vh" ) ) )
         switch( c ) {
             case 'v': debug_level++; break;
             case 'h':
             default : fprintf(stderr, "%s [<опции>] <имя_вх._файла.spx> <имя_вых._файла.wav>\n"
                                       "опции:\n"
                                       " -v - увеличить уровень детализации отладочного вывода\n"
                                       " -h - вот этот текст подсказки\n",
                               argv[ 0 ] );
                 exit( 'h' == c ? 0 : 1 );
         }
         

     

/*   if (strcmp(inFile, "-")==0)
   {
#if defined WIN32 || defined _WIN32
         _setmode(_fileno(stdin), _O_BINARY);
#endif
      fin=stdin;
   }
   else 
   {
      fin = fopen(inFile, "rb");
      if (!fin)
      {
         perror(inFile);
         exit(1);
      }
      close_in=1;
   } */
         
    int inpip[ 2 ], oupip[ 2 ];
    if( optind != argc - 2 ) fprintf(stderr, "No filenames specified\n"), exit(1);     
    // файл с которого читается входной (SPX) аудиопоток
    if ( !(c = open( argv[ optind ], O_RDONLY, S_IREAD ) ) ) perror("input open"), exit(1); 
    pipe( inpip );
    close( inpip[1] );    
    if ( -1 == dup2( c, inpip[0] ) ) perror("dup2"), exit(1);
    close( c );
    FILE* fin = fdopen( inpip[0], "rb" );
    if( NULL == fin ) perror("input reopen"), exit(1); 
    // новый файл в который пишется выходной (WAV) аудиопоток    
    if ( !(c = open( argv[ optind + 1 ], O_WRONLY | O_CREAT | O_TRUNC,
                     S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP | S_IROTH | S_IWOTH ) ) );
    pipe( oupip );
    close( oupip[0] );
    if ( -1 == dup2( c, oupip[1] ) ) perror("dup2"), exit(1);
    close( c );
    FILE* fou = fdopen( oupip[1], "wb" );
    if( NULL == fou ) perror("output reopen"), exit(1); 
//    decoder( fin, fou, debug_level == 0 );     
    encoder( fin, fou, debug_level == 0 );     
    fclose( fin );
    fclose( fou );
    exit( 0 );
}

//============================================================================
  


   /*Process command-line options
   while(1)
   {
      c = getopt_long (argc, argv, "nwuhvV",
                       long_options, &option_index);
      if (c==-1)
         break;
      
      switch(c)
      {
      case 0:
         if (strcmp(long_options[option_index].name,"narrowband")==0)
         {
            modeID = SPEEX_MODEID_NB;
         } else if (strcmp(long_options[option_index].name,"wideband")==0)
         {
            modeID = SPEEX_MODEID_WB;
         } else if (strcmp(long_options[option_index].name,"ultra-wideband")==0)
         {
            modeID = SPEEX_MODEID_UWB;
         } else if (strcmp(long_options[option_index].name,"vbr")==0)
         {
            vbr_enabled=1;
         } else if (strcmp(long_options[option_index].name,"abr")==0)
         {
            abr_enabled=atoi(optarg);
            if (!abr_enabled)
            {
               fprintf (stderr, "Invalid ABR value: %d\n", abr_enabled);
               exit(1);
            }
         } else if (strcmp(long_options[option_index].name,"vad")==0)
         {
            vad_enabled=1;
         } else if (strcmp(long_options[option_index].name,"dtx")==0)
         {
            dtx_enabled=1;
         } else if (strcmp(long_options[option_index].name,"quality")==0)
         {
            quality = atoi (optarg);
            vbr_quality=atof(optarg);
         } else if (strcmp(long_options[option_index].name,"bitrate")==0)
         {
            bitrate = atoi (optarg);
         } else if (strcmp(long_options[option_index].name,"nframes")==0)
         {
            nframes = atoi (optarg);
            if (nframes<1)
               nframes=1;
            if (nframes>10)
               nframes=10;
         } else if (strcmp(long_options[option_index].name,"comp")==0)
         {
            complexity = atoi (optarg);
         } else if (strcmp(long_options[option_index].name,"help")==0)
         {
            usage();
            exit(0);
         } else if (strcmp(long_options[option_index].name,"quiet")==0)
         {
            quiet = 1;
         } else if (strcmp(long_options[option_index].name,"version")==0)
         {
            version();
            exit(0);
         } else if (strcmp(long_options[option_index].name,"version-short")==0)
         {
            version_short();
            exit(0);
         } else if (strcmp(long_options[option_index].name,"le")==0)
         {
            lsb=1;
         } else if (strcmp(long_options[option_index].name,"be")==0)
         {
            lsb=0;
         } else if (strcmp(long_options[option_index].name,"8bit")==0)
         {
            fmt=8;
         } else if (strcmp(long_options[option_index].name,"16bit")==0)
         {
            fmt=16;
         } else if (strcmp(long_options[option_index].name,"stereo")==0)
         {
            chan=2;
         } else if (strcmp(long_options[option_index].name,"rate")==0)
         {
            rate=atoi (optarg);
         } else if (strcmp(long_options[option_index].name,"comment")==0)
         {
	   if (!strchr(optarg, '='))
	   {
	     fprintf (stderr, "Invalid comment: %s\n", optarg);
	     fprintf (stderr, "Comments must be of the form name=value\n");
	     exit(1);
	   }
           comment_add(&comments, &comments_length, NULL, optarg); 
         } else if (strcmp(long_options[option_index].name,"author")==0)
         {
           comment_add(&comments, &comments_length, "author=", optarg); 
         } else if (strcmp(long_options[option_index].name,"title")==0)
         {
           comment_add(&comments, &comments_length, "title=", optarg); 
         }

         break;
      case 'n':
         modeID = SPEEX_MODEID_NB;
         break;
      case 'h':
         usage();
         exit(0);
         break;
      case 'v':
         version();
         exit(0);
         break;
      case 'V':
         print_bitrate=1;
         break;
      case 'w':
         modeID = SPEEX_MODEID_WB;
         break;
      case 'u':
         modeID = SPEEX_MODEID_UWB;
         break;
      case '?':
         usage();
         exit(1);
         break;
      }
   }
   if (argc-optind!=2)
   {
      usage();
      exit(1);
   }
   inFile=argv[optind];
   outFile=argv[optind+1];



*/

