#include "common.h"

int read_wav_header(FILE *file, int *rate, int *channels, int *format, int *size) {
    char ch[5];
    int itmp;
    short stmp;
    int bpersec;
    short balign;
    int skip_bytes;
    int i;

   ch[4]=0;
#if 0
   fread(ch, 1, 4, file);
   if (strcmp(ch, "RIFF")!=0)
   {
      fseek(file, 0, SEEK_SET);
      return 0;
   }

   fread(&itmp, 4, 1, file);
   *size = le_int(itmp-36);

   fread(ch, 1, 4, file);
   if (strcmp(ch, "WAVE")!=0)
   {
      fprintf (stderr, "RIFF file is not a WAVE file\n");
      return -1;
   }
#endif
   fread(ch, 1, 4, file);
   while (strcmp(ch, "fmt ")!=0)
   {
      fread(&itmp, 4, 1, file);
      itmp = le_int(itmp);
      /*fprintf (stderr, "skip=%d\n", itmp);*/
      /*strange way of seeking, but it works even for pipes*/
      for (i=0;i<itmp;i++)
         fgetc(file);
      /*fseek(file, itmp, SEEK_CUR);*/
      fread(ch, 1, 4, file);
      if (feof(file))
      {
         fprintf (stderr, "Corrupted WAVE file: no \"fmt \"\n");
         return -1;
      }
   }
   /*if (strcmp(ch, "fmt ")!=0)
   {
      fprintf (stderr, "Corrupted WAVE file: no \"fmt \"\n");
      return -1;
      }*/
   
   fread(&itmp, 4, 1, file);
   itmp = le_int(itmp);
   skip_bytes=itmp-16;
   /*fprintf (stderr, "skip=%d\n", skip_bytes);*/
   
   fread(&stmp, 2, 1, file);
   stmp = le_short(stmp);
   if (stmp!=1)
   {
      fprintf (stderr, "Only PCM encoding is supported\n");
      return -1;
   }

   fread(&stmp, 2, 1, file);
   stmp = le_short(stmp);
   *channels = stmp;
   
   if (stmp>2)
   {
      fprintf (stderr, "Only mono and (intensity) stereo supported\n");
      return -1;
   }

   fread(&itmp, 4, 1, file);
   itmp = le_int(itmp);
   *rate = itmp;
   if (*rate != 8000 && *rate != 16000 && *rate != 11025 && *rate != 22050 && *rate != 32000 && *rate != 44100 && *rate != 48000)
   {
      fprintf (stderr, "Only 8 kHz (narrowband) and 16 kHz (wideband) supported (plus 11.025 kHz and 22.05 kHz, but your mileage may vary)\n");
      return -1;
   }

   fread(&itmp, 4, 1, file);
   bpersec = le_int(itmp);

   fread(&stmp, 2, 1, file);
   balign = le_short(stmp);

   fread(&stmp, 2, 1, file);
   stmp = le_short(stmp);
   if (stmp!=16 && stmp!=8)
   {
      fprintf (stderr, "Only 8/16-bit linear supported\n");
      return -1;
   }
   *format=stmp;

   if (bpersec!=*rate**channels*stmp/8)
   {
      fprintf (stderr, "Corrupted header: ByteRate mismatch\n");
      return -1;
   }

   if (balign!=*channels*stmp/8)
   {
      fprintf (stderr, "Corrupted header: BlockAlign mismatch\n");
      return -1;
   }

   
   /*strange way of seeking, but it works even for pipes*/
   if (skip_bytes>0)
      for (i=0;i<skip_bytes;i++)
         fgetc(file);

   /*fseek(file, skip_bytes, SEEK_CUR);*/

   fread(ch, 1, 4, file);
   while (strcmp(ch, "data")!=0)
   {
      fread(&itmp, 4, 1, file);
      itmp = le_int(itmp);
      /*strange way of seeking, but it works even for pipes*/
      for (i=0;i<itmp;i++)
         fgetc(file);
      /*fseek(file, itmp, SEEK_CUR);*/
      fread(ch, 1, 4, file);
      if (feof(file))
      {
         fprintf (stderr, "Corrupted WAVE file: no \"data\"\n");
         return -1;
      }
   }

   /*Ignore this for now*/
   fread(&itmp, 4, 1, file);
   itmp = le_int(itmp);

   *size=itmp;

   return 1;
}

void write_wav_header(FILE *file, int rate, int channels, int format, int size) {
    int itmp;
    short stmp;
    fprintf (file, "RIFF");
    itmp = 0x7fffffff;
    fwrite(&itmp, 4, 1, file);
    fprintf (file, "WAVEfmt ");
    itmp = le_int(16);
    fwrite(&itmp, 4, 1, file);
    stmp = le_short(1);
    fwrite(&stmp, 2, 1, file);
    stmp = le_short(channels);
    fwrite(&stmp, 2, 1, file);
    itmp = le_int(rate);
    fwrite(&itmp, 4, 1, file);
    itmp = le_int(rate*channels*2);
    fwrite(&itmp, 4, 1, file);
    stmp = le_short(2*channels);
    fwrite(&stmp, 2, 1, file);
    stmp = le_short(16);
    fwrite(&stmp, 2, 1, file);
    fprintf (file, "data");
    itmp = le_int(0x7fffffff);
    fwrite(&itmp, 4, 1, file);
}

#define readint(buf, base) (((buf[base+3]<<24)&0xff000000)| \
                           ((buf[base+2]<<16)&0xff0000)| \
                           ((buf[base+1]<<8)&0xff00)| \
  	           	    (buf[base]&0xff))

void print_comments(char *comments, int length) {
     char *c=comments, *end;
     int len, i, nb_fields;
     if (length < 8) {
         fprintf (stderr, "Invalid/corrupted comments\n");
         return;
     }
     end = c + length;
     len = readint(c, 0);
     c += 4;
     if (c + len > end) {
         fprintf (stderr, "Invalid/corrupted comments\n");
         return;
     }
     fwrite(c, 1, len, stderr);
     c += len;
     fprintf (stderr, "\n");
     if (c + 4 > end) {
         fprintf (stderr, "Invalid/corrupted comments\n");
         return;
     }
     nb_fields = readint(c, 0);
     c += 4;
     for (i = 0; i < nb_fields; i++) {
         if (c + 4 > end) {
             fprintf (stderr, "Invalid/corrupted comments\n");
             return;
         }
         len = readint(c, 0);
         c += 4;
         if (c + len > end) {
             fprintf (stderr, "Invalid/corrupted comments\n");
             return;
         }
         fwrite(c, 1, len, stderr);
         c += len;
         fprintf (stderr, "\n");
     }
}


void *process_header(ogg_packet *op, int enh_enabled, int *frame_size, 
                     int *rate, int *nframes, int forceMode, int *channels,
                     SpeexStereoState *stereo, int *extra_headers, int quiet) {
     void *st;
     SpeexMode *mode;
     SpeexHeader *header;
     int modeID;
     SpeexCallback callback;
     header = speex_packet_to_header((char*)op->packet, op->bytes);
     if (!header) {
         fprintf (stderr, "Cannot read header\n");
         return NULL;
     }
     if (header->mode >= SPEEX_NB_MODES) {
         fprintf (stderr, "Mode number %d does not (yet/any longer) exist in this version\n", header->mode);
         return NULL;
     }
     modeID = header->mode;
     if (forceMode!=-1) modeID = forceMode;
     mode = speex_lib_get_mode(modeID);
     if (header->speex_version_id > 1) {
         fprintf (stderr, "This file was encoded with Speex bit-stream version %d, which I don't know how to decode\n", header->speex_version_id);
         return NULL;
     }
     if (mode->bitstream_version < header->mode_bitstream_version) {
         fprintf (stderr, "The file was encoded with a newer version of Speex. You need to upgrade in order to play it.\n");
         return NULL;
     }
     if (mode->bitstream_version > header->mode_bitstream_version) {
         fprintf (stderr, "The file was encoded with an older version of Speex. You would need to downgrade the version in order to play it.\n");
         return NULL;
     }
     st = speex_decoder_init(mode);
     if (!st) {
         fprintf (stderr, "Decoder initialization failed.\n");
         return NULL;
     }
     speex_decoder_ctl(st, SPEEX_SET_ENH, &enh_enabled);
     speex_decoder_ctl(st, SPEEX_GET_FRAME_SIZE, frame_size);
     if (!(*channels==1)) {
         callback.callback_id = SPEEX_INBAND_STEREO;
         callback.func = speex_std_stereo_request_handler;
         callback.data = stereo;
         speex_decoder_ctl(st, SPEEX_SET_HANDLER, &callback);
     }
     if (!*rate) *rate = header->rate;
     // Adjust rate if --force-* options are used
     if (forceMode!=-1) {
         if (header->mode < forceMode) *rate <<= (forceMode - header->mode);
         if (header->mode > forceMode) *rate >>= (header->mode - forceMode);
     }
     speex_decoder_ctl(st, SPEEX_SET_SAMPLING_RATE, rate);
     *nframes = header->frames_per_packet;
     if (*channels==-1) *channels = header->nb_channels;
     if (!quiet) {
         fprintf (stderr, "Decoding %d Hz audio using %s mode", *rate, mode->modeName);
         if (*channels==1) fprintf (stderr, " (mono");
         else fprintf (stderr, " (stereo");
         if (header->vbr) fprintf (stderr, ", VBR)\n");
         else fprintf(stderr, ")\n");
         //fprintf (stderr, "Decoding %d Hz audio at %d bps using %s mode\n", 
         // *rate, mode->bitrate, mode->modeName);
     }
     *extra_headers = header->extra_headers;
     free(header);
     return st;
}

/*======================================================================================                 
 Comments will be stored in the Vorbis style.            
 It is describled in the "Structure" section of http://www.xiph.org/ogg/vorbis/doc/v-comment.html

The comment header is decoded as follows:
  1) [vendor_length] = read an unsigned integer of 32 bits
  2) [vendor_string] = read a UTF-8 vector as [vendor_length] octets
  3) [user_comment_list_length] = read an unsigned integer of 32 bits
  4) iterate [user_comment_list_length] times {
     5) [length] = read an unsigned integer of 32 bits
     6) this iteration's user comment = read a UTF-8 vector as [length] octets
     }
  7) [framing_bit] = read a single bit as boolean
  8) if ( [framing_bit]  unset or end of packet ) then ERROR
  9) done.

  If you have troubles, please write to ymnk@jcraft.com.
=======================================================================================*/

/*
#define readint(buf, base) (((buf[base+3]<<24)&0xff000000)| \
                           ((buf[base+2]<<16)&0xff0000)| \
                           ((buf[base+1]<<8)&0xff00)| \
  	           	    (buf[base]&0xff)) */

#define writeint(buf, base, val) do{ buf[base+3]=((val)>>24)&0xff; \
                                     buf[base+2]=((val)>>16)&0xff; \
                                     buf[base+1]=((val)>>8)&0xff; \
                                     buf[base]=(val)&0xff; \
                                 }while(0)

void comment_init(char **comments, int* length, char *vendor_string)
{
  int vendor_length=strlen(vendor_string);
  int user_comment_list_length=0;
  int len=4+vendor_length+4;
  char *p=(char*)malloc(len);
  if(p==NULL){
  }
  writeint(p, 0, vendor_length);
  memcpy(p+4, vendor_string, vendor_length);
  writeint(p, 4+vendor_length, user_comment_list_length);
  *length=len;
  *comments=p;
}

void comment_add(char **comments, int* length, char *tag, char *val)
{
  char* p = *comments;
  int vendor_length=readint(p, 0);
  int user_comment_list_length=readint(p, 4 + vendor_length);
  int tag_len=(tag?strlen(tag):0);
  int val_len=strlen(val);
  int len = (*length) + 4 + tag_len + val_len;

  p=(char*)realloc(p, len);
  if(p==NULL){
  }

  writeint(p, *length, tag_len+val_len);      // length of comment 
  if(tag) memcpy(p+*length+4, tag, tag_len);  // comment 
  memcpy(p+*length+4+tag_len, val, val_len);  // comment 
  writeint(p, 4+vendor_length, user_comment_list_length+1);

  *comments=p;
  *length=len;
}
#undef readint
#undef writeint

