#include "common.h"

static int debug_level = 0;

void decoder( FILE* fin, FILE* fout, int quiet ) { 
#define MAX_FRAME_SIZE 2000
     int option_index = 0;
     short out[MAX_FRAME_SIZE];
     short output[MAX_FRAME_SIZE];
     int frame_size=0;
     void *st=NULL;
     int packet_count=0;
     int stream_init = 0;
     ogg_int64_t page_granule=0, last_granule=0;
     int skip_samples=0, page_nb_packets;
     ogg_page       og;
     ogg_packet     op;
     ogg_stream_state os;
     int enh_enabled = 1;
     int nframes=2;
     int print_bitrate=0;
     int eos=0;
     int forceMode=-1;
     int audio_size=0;
     float loss_percent=-1;
     SpeexStereoState stereo = SPEEX_STEREO_STATE_INIT;
     int channels=-1;
     int rate=0;
     int extra_headers;
     int wav_format=0;
     int count = 0;

     //Init Ogg data struct
     ogg_sync_state oy;   
     ogg_sync_init(&oy);
     SpeexBits bits;   
     speex_bits_init(&bits);
     int nloop = 0, nsum = 0;
     //Main decoding loop
     while (1) {
         char *data;
         int i = 0, j, nb_read;
         nloop++;
         //Get the ogg buffer for writing
         data = ogg_sync_buffer(&oy, 200);
         //Read bitstream from input file
         nb_read = fread(data, sizeof(char), 200, fin);               
         if(debug_level > 0 ) fprintf (stderr, "loop #%d read from input %d\t=\t%d bytes\n", nloop, nb_read, nsum += nb_read);
         ogg_sync_wrote(&oy, nb_read);
         //Loop for all complete pages we got (most likely only one)
         while (ogg_sync_pageout(&oy, &og) == 1) {
             int packet_no;
             if (stream_init == 0) {
                 ogg_stream_init(&os, ogg_page_serialno(&og));
                 stream_init = 1;
             }
             //Add page to the bitstream
             ogg_stream_pagein(&os, &og);
             page_granule = ogg_page_granulepos(&og);
             page_nb_packets = ogg_page_packets(&og);
             if (page_granule>0 && frame_size) {
                 skip_samples = page_nb_packets*frame_size*nframes - (page_granule-last_granule);
                 if (ogg_page_eos(&og)) skip_samples = -skip_samples;
             //  else if (!ogg_page_bos(&og)) skip_samples = 0;
             } else skip_samples = 0;
             //printf ("page granulepos: %d %d %d\n", skip_samples, page_nb_packets, (int)page_granule);
             last_granule = page_granule;
             //Extract all available packets
             packet_no=0;
             while (!eos && ogg_stream_packetout(&os, &op)==1) {
                 if (packet_count==0) {
                     //If first packet, process as Speex header                 
                     st = process_header(&op, enh_enabled, &frame_size, &rate, &nframes, forceMode, &channels, &stereo, &extra_headers, quiet);
                     if (!nframes) nframes=1;
                     if (!st) exit(1);
   		     write_wav_header(fout, rate, channels, 0, 0);
                 } else if (packet_count==1) { //!!!!!!!!!!!!!!!!!!!!!
                     if (!quiet) print_comments((char*)op.packet, op.bytes);
                 } else if (packet_count<=1+extra_headers) { // Ignore extra headers 
                 } else {
                     int lost=0;
                     packet_no++;
                     if (loss_percent>0 && 100*((float)rand())/RAND_MAX<loss_percent) lost=1;
                     //End of stream condition
                     if (op.e_o_s) eos=1;
                     //Copy Ogg packet to Speex bitstream
                     speex_bits_read_from(&bits, (char*)op.packet, op.bytes);
                     for (j=0;j!=nframes;j++) {
                         int ret;
                         //Decode frame
                         if (!lost) ret = speex_decode_int(st, &bits, output);
                         else ret = speex_decode_int(st, NULL, output);
                         // for (i=0;i<frame_size*channels;i++) printf ("%d\n", (int)output[i]);
                         if (ret==-1) break;
                         if (ret==-2) {
                             fprintf (stderr, "Decoding error: corrupted stream?\n");
                             break;
                         }
                         if (speex_bits_remaining(&bits)<0) {
                             fprintf (stderr, "Decoding overflow: corrupted stream?\n");
                             break;
                         }
                         if (channels==2) speex_decode_stereo_int(output, frame_size, &stereo);
                         if (print_bitrate) {
                             int tmp;
                             char ch=13;
                             speex_decoder_ctl(st, SPEEX_GET_BITRATE, &tmp);
                             fputc (ch, stderr);
                             fprintf (stderr, "Bitrate is use: %d bps     ", tmp);
                         }
                         // Save to output file
                         for (i=0;i<frame_size*channels;i++) out[i]=output[i];
                         int frame_offset = 0;
                         int new_frame_size = frame_size;
                         //printf ("packet %d %d\n", packet_no, skip_samples);
                         if (packet_no == 1 && j==0 && skip_samples > 0) {
                             //printf ("chopping first packet\n");
                             new_frame_size -= skip_samples;
                             frame_offset = skip_samples;
                         }
                         if (packet_no == page_nb_packets && skip_samples < 0) {
                             int packet_length = nframes*frame_size+skip_samples;
                             new_frame_size = packet_length - j*frame_size;
                             if (new_frame_size<0) new_frame_size = 0;
                             if (new_frame_size>frame_size) new_frame_size = frame_size;
                             //printf ("chopping end: %d %d %d\n", new_frame_size, packet_length, packet_no);
                         }
                         if (new_frame_size) {  
                             fwrite(out+frame_offset*channels, sizeof(short), new_frame_size*channels, fout);
                             if(debug_level > 0 )
                                 fprintf (stderr, "write to array %d - frame_size %d - write to output file %d\n",
                                                  i, frame_size, (2*new_frame_size*channels));
                             audio_size += sizeof(short) * new_frame_size * channels;
                         }
                     }
                 }
                 packet_count++;
             }
         }
         if (feof(fin)) break;
     }
     if (wav_format) { // ??????????????????????
         if (fseek(fout, 4, SEEK_SET) == 0) {
             int tmp = le_int(audio_size + 36);
             fwrite(&tmp, 4, 1, fout);
             if (fseek(fout, 32, SEEK_CUR) == 0) {
                 tmp = le_int(audio_size);
                 fwrite(&tmp, 4, 1, fout);
             } else fprintf (stderr, "First seek worked, second didn't\n");
         } else fprintf (stderr, "Cannot seek on wave file, size will be incorrect\n");
     }
     if (st) speex_decoder_destroy(st);
     else fprintf (stderr, "This doesn't look like a Speex file\n");
     speex_bits_destroy(&bits);
     if (stream_init) ogg_stream_clear(&os);
     ogg_sync_clear(&oy);
}

int main(int argc, char **argv) {
     int c;
     while( -1 != ( c = getopt( argc, argv, "vh" ) ) )
         switch( c ) {
             case 'v': debug_level++; break;
             case 'h':
             default : fprintf(stderr, "%s [<опции>] <имя_вх._файла.spx> <имя_вых._файла.wav>\n"
                                       "опции:\n"
                                       " -v - увеличить уровень детализации отладочного вывода\n"
                                       " -h - вот этот текст подсказки\n",
                               argv[ 0 ] );
                 exit( 'h' == c ? 0 : 1 );
         }
    int inpip[ 2 ], oupip[ 2 ];
    if( optind != argc - 2 ) fprintf(stderr, "No filenames specified\n"), exit(1);     
    // файл с которого читается входной (SPX) аудиопоток
    if ( !(c = open( argv[ optind ], O_RDONLY, S_IREAD ) ) ) perror("input open"), exit(1); 
    pipe( inpip );
    close( inpip[1] );    
    if ( -1 == dup2( c, inpip[0] ) ) perror("dup2"), exit(1);
    close( c );
    FILE* fin = fdopen( inpip[0], "rb" );
    if( NULL == fin ) perror("input reopen"), exit(1); 
    // новый файл в который пишется выходной (WAV) аудиопоток    
    if ( !(c = open( argv[ optind + 1 ], O_WRONLY | O_CREAT | O_TRUNC,
                     S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP | S_IROTH | S_IWOTH ) ) );
    pipe( oupip );
    close( oupip[0] );
    if ( -1 == dup2( c, oupip[1] ) ) perror("dup2"), exit(1);
    close( c );
    FILE* fou = fdopen( oupip[1], "wb" );
    if( NULL == fou ) perror("output reopen"), exit(1); 
    decoder( fin, fou, debug_level == 0 );     
    fclose( fin );
    fclose( fou );
    exit( 0 );
}

