#include "common.h"

#define MAX_FRAME_SIZE 2000
#define MAX_FRAME_BYTES 2000

void decoder( FILE* fin, FILE* fout, int quiet ) {
   short out[MAX_FRAME_SIZE];
   short output[MAX_FRAME_SIZE];
   int frame_size = 0;
   int packet_count = 0;
   int stream_init = 0;
   ogg_int64_t page_granule = 0, last_granule = 0;
   int skip_samples = 0, page_nb_packets;
   ogg_page       og;
   ogg_packet     op;
   ogg_stream_state os;
   int nframes = 2;
   int audio_size = 0;
   SpeexStereoState stereo = SPEEX_STEREO_STATE_INIT;
   int channels = -1;
   int rate = 0;
   int extra_headers;
   int count = 0;

   //Init Ogg data struct
   ogg_sync_state oy;
   ogg_sync_init(&oy);
   SpeexBits bits;
   speex_bits_init(&bits);
   int nloop = 0, nsum = 0;
   void *st = NULL;
   //Main decoding loop
   int eos = 0;
   while (1) {
      char *data;
      int i = 0, j, nb_read;
      nloop++;
      //Get the ogg buffer for writing
      data = ogg_sync_buffer(&oy, 200);
      //Read bitstream from input file
      nb_read = fread(data, sizeof(char), 200, fin);
      if (quiet > 1) fprintf (stderr, "#%d\tread from input %d\t=\t%d bytes\n", nloop, nb_read, nsum += nb_read);
      ogg_sync_wrote(&oy, nb_read);
      //Loop for all complete pages we got (most likely only one)
      while (ogg_sync_pageout(&oy, &og) == 1) {
         int packet_no;
         if (stream_init == 0) {
            ogg_stream_init(&os, ogg_page_serialno(&og));
            stream_init = 1;
         }
         //Add page to the bitstream
         ogg_stream_pagein(&os, &og);
         page_granule = ogg_page_granulepos(&og);
         page_nb_packets = ogg_page_packets(&og);
         if (page_granule > 0 && frame_size) {
            skip_samples = page_nb_packets * frame_size * nframes - (page_granule - last_granule);
            if (ogg_page_eos(&og)) skip_samples = -skip_samples;
            //  else if (!ogg_page_bos(&og)) skip_samples = 0;
         } else skip_samples = 0;
         //printf ("page granulepos: %d %d %d\n", skip_samples, page_nb_packets, (int)page_granule);
         last_granule = page_granule;
         //Extract all available packets
         packet_no=0;
         while (!eos && ogg_stream_packetout(&os, &op) == 1) {
            if (packet_count==0) {
               const int forceMode = -1;
               //If first packet, process as Speex header
               st = process_header(&op, 1, &frame_size, &rate, &nframes, -1, &channels, &stereo, &extra_headers, quiet);
               if (!nframes) nframes=1;
               if (!st) exit(1);
               write_wav_header(fout, rate, channels, 0, 0);
            } else if (packet_count==1) {
               if (!quiet) print_comments((char*)op.packet, op.bytes);
            } else if (packet_count <= 1 + extra_headers) { // Ignore extra headers
            } else {
               packet_no++;
               //End of stream condition
               if (op.e_o_s) eos = 1;
               //Copy Ogg packet to Speex bitstream
               speex_bits_read_from(&bits, (char*)op.packet, op.bytes);
               for (j = 0; j != nframes; j++) {
                  //Decode frame
                  int ret = speex_decode_int(st, &bits, output);
                  // for (i=0;i<frame_size*channels;i++) printf ("%d\n", (int)output[i]);
                  if (ret == -1) break;
                  if (ret == -2) {
                     fprintf (stderr, "Decoding error: corrupted stream?\n");
                     break;
                  }
                  if (speex_bits_remaining(&bits) < 0) {
                     fprintf (stderr, "Decoding overflow: corrupted stream?\n");
                     break;
                  }
                  if (channels == 2) speex_decode_stereo_int(output, frame_size, &stereo);
                  // Save to output file
                  for (i = 0; i < frame_size * channels; i++) out[i] = output[i];
                  int frame_offset = 0;
                  int new_frame_size = frame_size;
                  //printf ("packet %d %d\n", packet_no, skip_samples);
                  if (packet_no == 1 && j == 0 && skip_samples > 0) {
                     //printf ("chopping first packet\n");
                     new_frame_size -= skip_samples;
                     frame_offset = skip_samples;
                  }
                  if (packet_no == page_nb_packets && skip_samples < 0) {
                     int packet_length = nframes * frame_size + skip_samples;
                     new_frame_size = packet_length - j * frame_size;
                     if (new_frame_size < 0) new_frame_size = 0;
                     if (new_frame_size > frame_size) new_frame_size = frame_size;
                     //printf ("chopping end: %d %d %d\n", new_frame_size, packet_length, packet_no);
                  }
                  if (new_frame_size) {
                	 int samples = new_frame_size * channels;
                     fwrite(out + frame_offset * channels, sizeof(short), samples, fout);
                     audio_size += sizeof(short) * samples;
                     if (quiet > 1)
                        fprintf (stderr,  "#%d\twrite to output %d\t=\t%d bytes\n",
                                 nloop, sizeof(short) * samples, audio_size);
                  }
               }
            }
            packet_count++;
         }
      }
      if (feof(fin)) break;
   }
   if (fseek(fout, 4, SEEK_SET) == 0) {
      int tmp = le_int(audio_size + 36);
      fwrite(&tmp, 4, 1, fout);
      if (fseek(fout, 32, SEEK_CUR) == 0) {
         tmp = le_int(audio_size);
         fwrite(&tmp, 4, 1, fout);
      } else fprintf (stderr, "First seek worked, second didn't\n");
   } else fprintf (stderr, "Cannot seek on wave file, size will be incorrect\n");
   if (st) speex_decoder_destroy(st);
   else fprintf (stderr, "This doesn't look like a Speex file\n");
   speex_bits_destroy(&bits);
   if (stream_init) ogg_stream_clear(&os);
   ogg_sync_clear(&oy);
}

//Write an Ogg page to a file pointer
static int oe_write_page(ogg_page *page, FILE *fp) {
   int written;
   written = fwrite(page->header,1,page->header_len, fp);
   written += fwrite(page->body,1,page->body_len, fp);
   return written;
}

// Convert input audio bits, endians and channels
static int read_samples(FILE *fin, int frame_size, int bits, int channels, int lsb, short * input, char *buff, int *size) {
   unsigned char in[MAX_FRAME_BYTES * 2];
   int i;
   short *s;
   int nb_read;
   if (size && *size <= 0) return 0;
   //Read input audio
   if (size) *size -= bits / 8 * channels * frame_size;
   if (buff) {
      for (i = 0; i < 12; i++) in[i] = buff[i];
      nb_read = fread(in+12, 1, bits / 8 * channels * frame_size - 12, fin) + 12;
      if (size) *size += 12;
   } else nb_read = fread(in,1,bits/8*channels* frame_size, fin);
   nb_read /= bits / 8 * channels;
   //fprintf (stderr, "%d\n", nb_read);
   if (nb_read == 0) return 0;
   s = (short*)in;
   if(bits == 8)
      // Convert 8->16 bits
      for(i = frame_size * channels - 1; i >= 0; i-- ) s[i] = (in[i] << 8) ^ 0x8000;
   else
      // convert to our endian format
      for(i = 0; i < frame_size * channels; i++)
         if(lsb) s[i] = le_short(s[i]);
         else s[i] = be_short(s[i]);
   // FIXME: This is probably redundent now copy to float input buffer
   for (i = 0; i < frame_size * channels; i++) input[i]=(short)s[i];
   for (i = nb_read * channels; i < frame_size * channels; i++) input[i] = 0;
   return nb_read;
}

#define SPEEX_VERSION "speex-1.0.4"
static char *vendor_string = "Encoded with Speex " SPEEX_VERSION;

void encoder( FILE* fin, FILE* fout, int quiet ) {
   int nb_samples, total_samples=0, nb_encoded;
   int size;
   int bytes_written=0, ret, result;
   int modeID = SPEEX_MODEID_NB, rate=8000, chan=1, fmt=16;
   char first_bytes[12];
   fread(first_bytes, 1, 12, fin);
   if (strncmp(first_bytes,"RIFF",4) == 0 && strncmp(first_bytes,"RIFF",4) == 0) {
         if (read_wav_header(fin, &rate, &chan, &fmt, &size) == -1)
        	 fprintf(stderr,"Error: wav file header\n"), exit(1);
   }
   else fprintf(stderr,"Error: wav file header\n"), exit(1);
   SpeexMode *mode = speex_lib_get_mode(modeID);
   SpeexHeader header;
   speex_init_header(&header, rate, 1, mode);
   const int nframes = 1;
   header.frames_per_packet = nframes;
   header.vbr = 0;
   header.nb_channels = chan;
   if (quiet > 0) fprintf (stderr, "Encoding %d Hz audio using %s mode (%s)\n",
                           header.rate, mode->modeName, chan==2 ? "stereo" : "mono");
   //Initialize Speex encoder
   void *st = speex_encoder_init(mode);
   int frame_size;
   speex_encoder_ctl(st, SPEEX_GET_FRAME_SIZE, &frame_size);
   int complexity = 3;
   speex_encoder_ctl(st, SPEEX_SET_COMPLEXITY, &complexity);
   speex_encoder_ctl(st, SPEEX_SET_SAMPLING_RATE, &rate);
   int quality = -1;
   if (quality >= 0) speex_encoder_ctl(st, SPEEX_SET_QUALITY, &quality);
   int lookahead = 0;
   speex_encoder_ctl(st, SPEEX_GET_LOOKAHEAD, &lookahead);
   //Initialize Ogg stream struct
   ogg_stream_state os;
   srand(time(NULL));
   if (ogg_stream_init(&os, rand()) == -1) fprintf(stderr,"Error: ogg stream init failed\n"), exit(1);
   //Write header (with comments)
   ogg_packet   op;
   op.packet = (unsigned char *)speex_header_to_packet(&header, (int*)&(op.bytes));
   op.b_o_s = 1;
   op.e_o_s = 0;
   op.granulepos = 0;
   op.packetno = 0;
   ogg_stream_packetin(&os, &op);
   //Initialize comments (vendor)
   char *comments;
   int comments_length;
   comment_init(&comments, &comments_length, vendor_string);
   free(op.packet);
   op.packet = (unsigned char *)comments;
   op.bytes = comments_length;
   op.b_o_s = 0;
   op.e_o_s = 0;
   op.granulepos = 0;
   op.packetno = 1;
   ogg_stream_packetin(&os, &op);
   ogg_page 	    og;
   while((result = ogg_stream_flush(&os, &og))) {
      if(!result) break;
      ret = oe_write_page(&og, fout);
      if(ret != og.header_len + og.body_len) fprintf (stderr,"Error: failed writing header to output stream\n"), exit(1);
      else bytes_written += ret;
   }
   free(comments);
   SpeexBits bits;
   speex_bits_init(&bits);
   short input[MAX_FRAME_SIZE];
   int lsb=1;
   nb_samples = read_samples(fin, frame_size, fmt, chan, lsb, input, NULL, &size);
   int eos = 0;
   if (nb_samples == 0) eos = 1;
   total_samples += nb_samples;
   int id = -1;
   if (quiet > 1)
      fprintf (stderr, "#%d\tread from input %d\t=\t%d bytes\n", id + 1, nb_samples, total_samples);
   nb_encoded = -lookahead;
   //Main encoding loop (one frame per iteration)
   while (!eos || total_samples > nb_encoded) {
      char cbits[MAX_FRAME_BYTES];
      int nbBytes;
      id++;
      //Encode current frame
      if (chan==2) speex_encode_stereo_int(input, frame_size, &bits);
      speex_encode_int(st, input, &bits);
      nb_encoded += frame_size;
      nb_samples = read_samples(fin, frame_size, fmt, chan, lsb, input, NULL, &size);
      if (nb_samples == 0) eos = 1;
      if (eos && total_samples<=nb_encoded) op.e_o_s = 1;
      else op.e_o_s = 0;
      total_samples += nb_samples;
      if (quiet > 1)
         fprintf (stderr, "#%d\tread from input %d\t=\t%d bytes\n", id + 1, nb_samples, total_samples);
      speex_bits_insert_terminator(&bits);
      nbBytes = speex_bits_write(&bits, cbits, MAX_FRAME_BYTES);
      speex_bits_reset(&bits);
      op.packet = (unsigned char *)cbits;
      op.bytes = nbBytes;
      op.b_o_s = 0;
      //Is this redundent?
      if (eos && total_samples <= nb_encoded) op.e_o_s = 1; else op.e_o_s = 0;
      op.granulepos = (id + 1) * frame_size - lookahead;
      if (op.granulepos > total_samples) op.granulepos = total_samples;
      /*printf ("granulepos: %d %d %d %d %d %d\n", (int)op.granulepos, id, nframes, lookahead, 5, 6);*/
      op.packetno = 2 + id / nframes;
      ogg_stream_packetin(&os, &op);
      //Write all new pages (most likely 0 or 1)
      while (ogg_stream_pageout(&os,&og)) {
         ret = oe_write_page(&og, fout);
         if(ret != og.header_len + og.body_len) fprintf (stderr,"Error: failed writing header to output stream\n"), exit(1);
         else bytes_written += ret;
         if (quiet > 1)
            fprintf (stderr, "#%d\twrite to output %d\t=\t%d bytes\n", id + 1, bytes_written);
      }
   }
   //Flush all pages left to be written
   while (ogg_stream_flush(&os, &og)) {
      ret = oe_write_page(&og, fout);
      if(ret != og.header_len + og.body_len) fprintf (stderr, "Error: failed writing header to output stream\n"), exit(1);
      else bytes_written += ret;
      if (quiet > 1)
         fprintf (stderr, "#%d write to output %d\t=\t%d bytes\n", id + 2, bytes_written);
   }
   speex_encoder_destroy(st);
   speex_bits_destroy(&bits);
   ogg_stream_clear(&os);
   return;
}

int main(int argc, char **argv) {
    int c, enc = 0, debug_level = 0;
    while( -1 != ( c = getopt( argc, argv, "edvh" ) ) )
        switch( c ) {
            case 'e':
            case 'd': enc = 'e' == c ? 1 : -1; break;
            case 'v': debug_level++; break;
            case 'h':
            default : fprintf(stderr, "%s [<опции>] <имя_вх._файла.spx> <имя_вых._файла.wav>\n"
                                      "опции:\n"
            		                  " -e - кодировать WAV - > SPEEX\n"
            		                  " -d - декодировать SPEEX - > WAV\n"
                                      " -v - увеличить уровень детализации отладочного вывода\n"
                                      " -h - вот этот текст подсказки\n",
                              argv[ 0 ] );
                exit( 'h' == c ? 0 : 1 );
        }
    if( optind != argc - 2 ) fprintf(stderr, "Должно быть указано 2 файла\n"), exit(1);
    if( 0 == enc ) fprintf(stderr, "Не указано направление преобразования: '-e' | '-d'\n"), exit(1);
    int inpip[ 2 ], oupip[ 2 ];
    // файл с которого читается входной (SPX) аудиопоток
    if ( !(c = open( argv[ optind ], O_RDONLY, S_IREAD ) ) ) perror("input open"), exit(1);
    pipe( inpip );
    close( inpip[1] );
    if ( -1 == dup2( c, inpip[0] ) ) perror("dup2"), exit(1);
    close( c );
    FILE* fin = fdopen( inpip[0], "rb" );
    if( NULL == fin ) perror("input reopen"), exit(1);
    // новый файл в который пишется выходной (WAV) аудиопоток
    if ( !(c = open( argv[ optind + 1 ], O_WRONLY | O_CREAT | O_TRUNC,
                     S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP | S_IROTH | S_IWOTH ) ) );
    pipe( oupip );
    close( oupip[0] );
    if ( -1 == dup2( c, oupip[1] ) ) perror("dup2"), exit(1);
    close( c );
    FILE* fou = fdopen( oupip[1], "wb" );
    if( NULL == fou ) perror("output reopen"), exit(1);
    if( enc < 0 ) decoder( fin, fou, debug_level );
    else encoder( fin, fou, debug_level );
    fclose( fin );
    fclose( fou );
    exit( 0 );
}
